import logging
import liblynx
from django.conf import settings
from django.http import HttpResponseRedirect


class LibLynxAuthMiddleware:
    def __init__(self, get_response):
        self.ll = liblynx.Connect()
        logging.debug(
            f"LibLynx API connected with access_token: {self.ll.access_token}"
        )
        self.get_response = get_response
        self.product_cache = {}

    def authenticate(self, request):
        host = request.META.get("HTTP_X_FORWARDED_FOR", "127.0.0.127")
        abs_uri = request.build_absolute_uri()
        user_agent = request.headers.get("User-Agent", "<unknown>")

        logging.debug(
            f"Trying New identification for \nhost: [{host}]\nurl: [{abs_uri}]\nuser agent: [{user_agent}]",
        )
        identification = self.ll.new_identification(host, abs_uri, user_agent,)
        logging.debug(identification)

        if (
            "_ll_allow_wayf" in request.GET and identification["status"] == "wayf"
        ):  # Needs a redirect to authenticate
            return HttpResponseRedirect(identification["_links"]["wayf"]["href"])

        if identification["status"] == "identified":
            request.session["LIBLYNX_ACCOUNT"] = identification["account"][
                "account_name"
            ]
            request.session["LIBLYNX_ACCOUNT_ID"] = identification["account"]["id"]

            # Get the account information, so that we can retrieve the subscriptions
            account = self.ll.api(
                "account", identification["account"]["_links"]["self"]["href"],
            )
            logging.debug(account)

            # Fetch the products for this user, if configured
            if settings.LIBLYNX_FETCH_PRODUCTS:
                subs = self.ll.api("subs", account["_links"]["account_subs"]["href"])

                products = []
                for sub in subs["subscriptions"]:
                    sub_link = sub["_links"]["subscription_units"]["href"]
                    if sub_link not in self.product_cache:
                        self.product_cache[sub_link] = self.ll.api("sub", sub_link)
                    for unit in self.product_cache[sub_link]["units"]:
                        products.append(unit["unit_code"])

                request.session["LIBLYNX_PRODUCTS"] = products
                logging.debug(f"Products found {products}")

    def __call__(self, request):
        if "LIBLYNX_ACCOUNT" not in request.session:
            response = self.authenticate(request)
            if response is not None:
                return response

        response = self.get_response(request)

        return response



import logging
import liblynx

from starlette.authentication import (
    AuthenticationBackend,
    AuthenticationError,
    SimpleUser,
    AuthCredentials,
)
from starlette.datastructures import URL
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import PlainTextResponse, RedirectResponse
from starlette.types import ASGIApp, Receive, Scope, Send

import hmac
import base64


class LibLynxUser(SimpleUser):
    def __init__(self, account_id: str, username: str) -> None:
        self.account_id = str(account_id)
        self.username = username


class LibLynxAuthHTTPMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)
        self.connect = liblynx.Connect()

    async def authenticate(self, request):
        llid = request.cookies.get("llid")
        if llid:
            logging.debug("Has cookie %s" % llid)
            try:
                plain_cookie = base64.b64decode(llid.encode("utf8")).decode("utf8")
                cookie_version, llid, ll_accountname, llid_sig = plain_cookie.split(
                    "\n"
                )
                sig = hmac.new(
                    self.connect.client_secret.encode("utf8"),
                    llid.encode("utf8"),
                    digestmod="sha256",
                ).hexdigest()
                logging.debug("sig %s" % sig)
                logging.debug("llid_sig %s" % llid_sig)
                if sig == llid_sig:
                    return (
                        AuthCredentials(["authenticated"]),
                        LibLynxUser(llid, ll_accountname),
                    )
            except:
                logging.debug("Cookie parsing in LibLynxAuthBackend caused exception",)
                raise AuthenticationError("Cookie Parsing failed")

        identification = self.connect.new_identification(
            request.headers.get("X-Forwarded-For", request.client.host),
            str(URL(scope=request.scope)),
            request.headers.get("User-Agent", "<unknown>"),
        )
        logging.debug(
            "identification in LibLynxAuthBackend is %s", identification["id"]
        )

        if identification["status"] == "identified":
            return (
                AuthCredentials(["authenticated"]),
                LibLynxUser(
                    identification["account"].get("publisher_reference")
                    or "brill",  # NOTE how we use the Publisher reference and NOT the account id
                    identification["account"]["account_name"],
                ),
            )
        return  (
                AuthCredentials(["authenticated"]),
                LibLynxUser(
                    "anon",  # NOTE how we use the Publisher reference and NOT the account id
                    "Anonymous",
                ),
            )

        # Disable WAYF temporarily, and make all access Anonymous if not directly identified
        # elif identification["status"] == "wayf":
        #     logging.debug(
        #         "WAYF needed, asking for %s", identification["_links"]["wayf"]["href"]
        #     )
        #     raise WAYFException(identification["_links"]["wayf"]["href"])
        return AuthCredentials(), UnauthenticatedUser()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = Request(scope, receive=receive)
        try:
            scope["auth"], scope["user"] = await self.authenticate(request)
            response = await self.dispatch_func(request, self.call_next)
        except AuthenticationError as exc:
            if isinstance(exc, WAYFException):
                response = RedirectResponse(url=exc.redirect_to)
            else:
                response = PlainTextResponse(str(exc), status_code=400)
                response.delete_cookie("llid")

        await response(scope, receive, send)

    async def dispatch(self, request, call_next):
        response = await call_next(request)
        if not request.cookies.get("llid") and isinstance(request.user, LibLynxUser):
            sig = hmac.new(
                self.connect.client_secret.encode("utf8"),
                request.user.account_id.encode("utf8"),
                digestmod="sha256",
            ).hexdigest()
            logging.debug("signed cookie sig %s" % sig)
            cookie = "2\n%s\n%s\n%s" % (
                request.user.account_id,
                request.user.username,
                sig,
            )
            cookie = base64.b64encode(cookie.encode("utf8")).decode("utf8")
            response.set_cookie(
                "llid", cookie, expires=(60 * 60 * 24 * 7)
            )  # expires in one week
        return response


class WAYFException(AuthenticationError):
    def __init__(self, redirect_to: str = None) -> None:
        self.redirect_to = redirect_to



__version__ = "0.12"
import logging
import os
import time
from email.utils import parsedate
import requests


def make_url(path):
    BASE = os.environ.get("LIBLYNX_BASE") or "https://sandbox.liblynx.com"
    return "%s%s" % (BASE, path)


class APICallException(Exception):
    pass


def if_success(arequest, msg, successcode=200):
    if arequest.status_code == successcode:
        logging.debug("%s worked" % msg)
        return arequest.json()
    else:
        raise APICallException(
            "%s failed. status: %s %s" % (msg, arequest.status_code, arequest.text)
        )


class Connect:
    def __init__(self, client_id=None, client_secret=None):
        client_id = client_id or os.environ.get("LIBLYNX_CLIENT_ID")
        if not client_id:
            raise Exception(
                "client_id not specified nor found from ENV LIBLYNX_CLIENT_ID"
            )
        self.client_id = client_id
        client_secret = client_secret or os.environ.get("LIBLYNX_CLIENT_SECRET")
        if not client_secret:
            raise Exception(
                "client_secret not specified nor found from ENV LIBLYNX_CLIENT_SECRET"
            )
        self.client_secret = client_secret
        self.CACHE = {}

    def __getattr__(self, attr):
        value, expiry = self.CACHE.get(attr, (None, 0))
        if time.time() > expiry:
            getter = "_get_%s" % attr
            logging.debug("%s expired, trying getattr with %s" % (attr, getter))
            getter = self.__getattribute__(getter)
            if getter:
                value, expiry = getter()
                logging.debug(
                    "got new value for %s expires at %s" % (attr, time.ctime(expiry))
                )
                self.CACHE[attr] = (value, expiry)
        return value

    def _get_access_token(self):
        r = requests.post(
            make_url("/oauth/v2/token"),
            {"grant_type": "client_credentials"},
            auth=(self.client_id, self.client_secret),
        )
        if r.status_code == 200:
            r_json = r.json()
            logging.debug("access_token acquired for %s", self.client_id)
            expiry = time.time() + (r_json["expires_in"] - 60)
            return r_json["access_token"], expiry
        logging.error("access_token failed %s %s", r.status_code, r.text)

    def _get_endpoint(self):
        logging.debug("api endpoint expired getting a new one")
        headers = {
            "Authorization": "Bearer %s" % self.access_token,
            "Accept": "application/json",
        }
        r = requests.get(make_url("/api"), headers=headers)
        if r.status_code == 200:
            # Make it one hour less than the expiry
            expires = time.mktime(parsedate((r.headers.get("Expires")))) - (60 * 60)
            api = r.json()
            logging.debug("new api endpoint acquired")
            return api, expires
        else:
            raise APICallException(
                "api endpoint failed %s %s" % (r.status_code, r.text)
            )

    def api(self, msg, href, method="GET", data=None, successcode=200):
        headers = {
            "Authorization": "Bearer %s" % self.access_token,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        if method == "GET":
            r = requests.get(href, headers=headers,)
        elif method == "POST":
            r = requests.post(href, json=data, headers=headers,)
        elif method == "DELETE":
            r = requests.delete(href, headers=headers,)
            return True
        return if_success(r, msg, successcode)

    def new_identification(self, ip, url, user_agent):
        return self.api(
            "new identification",
            self.endpoint["_links"]["@new_identification"]["href"],
            "POST",
            {"ip": ip, "url": url, "user_agent": user_agent},
            201,
        )

    def identification(self, anid):
        href = self.endpoint["_links"]["@get_identification"]["href"].replace(
            "{id}", anid
        )
        return self.api("get identification", href)


# c.api("new account", c.endpoint["_links"]["@new_account"]["href"], "POST", {"account_name":"LCI Evaluation %s", "subscriptions":[{"title":"Foo", "start": "2020-11-01 10:56:58", "end": "2020-12-01 10:56:58"}]})
